package de.uni_passau.fim.evaluator;

import java.util.Calendar;
import java.util.Date;
import java.util.List;

import de.uni_passau.fim.strategies.CompleteGuessIterator;
import de.uni_passau.fim.strategies.Configuration;
import de.uni_passau.fim.strategies.Guess;
import de.uni_passau.fim.strategies.IGuessIterator;
import de.uni_passau.fim.strategies.IStrategy;
import de.uni_passau.fim.strategies.Logger;
import de.uni_passau.fim.strategies.RandomGuessIterator;

public class MyEvaluator {

	private Mastermind mBoard;
	private IStrategy mStrategy;
	
	private Stats mStats;
	
	private IGuessIterator mGuessIterator;
	
	private BoardInfo mBoardInfo;
	
	public MyEvaluator(Mastermind m, IStrategy s, boolean randomGames, int numOfRandomGames) {
		if (randomGames)
			mGuessIterator = new RandomGuessIterator(m.getColors(), m.getPlaces(), m.isWithRepetition(), numOfRandomGames);
		else
			mGuessIterator = new CompleteGuessIterator(m.getColors(), m.getPlaces(), m.isWithRepetition());
		mStrategy = s;
		mBoard = m;
		setBoardInfo();
		mStats = new Stats();
		mStats.numOfGamesForStepNum = new int[m.getRows()+1];
		mStats.totalNumberOfGames = mGuessIterator.size();
	}
	
	private void setBoardInfo() {
		mBoardInfo = new BoardInfo();
		mBoardInfo.numOfColors = mBoard.getColors();
		mBoardInfo.numOfPlaces = mBoard.getPlaces();
		mBoardInfo.numOfRows = mBoard.getRows();
		mBoardInfo.withRepetition = mBoard.isWithRepetition();
	}
	
	public int remainingGames() {
		return totalNumberOfGames() - playedGames();
	}
	
	public int totalNumberOfGames() {
		return mStats.totalNumberOfGames;
	}
	
	public int playedGames() {
		return mStats.numberOfPlayedGames;
	}
	
	public boolean hasMoreGames() {
		return mStats.numberOfPlayedGames < mStats.totalNumberOfGames;
	}
	
	public String getStrategyInfo() {
		if (mStrategy == null)
			return "";
		else
			return mStrategy.info();
	}
	
	public Stats getStats() {
		return mStats;
	}
	
	public void playNext() {
		System.out.println("Playing next game...");
		if (!mGuessIterator.hasNext())
			return;
		mBoard.init(mGuessIterator.next());
		
		GameInfo info = play(mBoard, mStrategy);
		updateStats(info);
	}
	
	private void updateStats(GameInfo info) {
		mStats.numberOfPlayedGames++;
		mStats.totalTime += info.time;
		if (info.steps == -1) {
			mStats.numberOfResignedGames++;
		} else {
			mStats.numOfGamesForStepNum[info.steps]++;
			mStats.numberOfStepsTotal += info.steps;
		}
		if (info.solved)
			mStats.numberOfSolvedGames++;
		else if (info.steps != -1)
			mStats.numberOfUnsolvedGames++;
		if (info.time > mStats.maxTime)
			mStats.maxTime = info.time;
		mStats.totalTime += info.time;
	}

	/**
	 * Play a single game. Assumes that <code>m</code> already has been
	 * initialized. <code>s</code> will be resetted.
	 */
	private GameInfo play(Mastermind m, IStrategy s){
		IStrategy strategy = s;
		int steps = 0;
		s.reset();
		Date startOfGame = Calendar.getInstance().getTime();
		Logger.info(m.toString());
		while (!m.solved() && steps < m.getRows()) {
			List<Configuration> cs = m.getCurrentState();
			Date startOfGuess = Calendar.getInstance().getTime();
			Guess g = strategy.getGuess(cs);
			Date endOfGuess = Calendar.getInstance().getTime();
			float durationOfGuess = endOfGuess.getTime() - startOfGuess.getTime();
			if (g == null) {
				Date endOfGame = Calendar.getInstance().getTime();
				float durationOfGame = endOfGame.getTime() - startOfGame.getTime();
				return new GameInfo(durationOfGame, -1, false /* Not solved */);
			}
			m.guess(g);
			steps++;
			Logger.info(m.toString());
		}
		Date endOfGame = Calendar.getInstance().getTime();
		float durationOfGame = endOfGame.getTime() - startOfGame.getTime();
		return new GameInfo(durationOfGame, steps, m.solved());
	}
	
	public static class Stats {
		public int totalNumberOfGames = 0;
		public int numberOfSolvedGames = 0;
		public int numberOfPlayedGames = 0;
		public int numberOfResignedGames = 0;
		public int numberOfUnsolvedGames = 0;
		public int numberOfStepsTotal = 0;
		public float maxTime = -1;
		public float totalTime = 0;
		public int[] numOfGamesForStepNum;
		
		public String toString() {
			StringBuffer s = new StringBuffer();
			s.append("Results:\n");
			s.append("# of played games: " + this.numberOfPlayedGames);
			s.append("\n");
			s.append("# of solved games: " + this.numberOfSolvedGames);
			s.append("\n");
			s.append("# of unsolved games: " + this.numberOfUnsolvedGames);
			s.append("\n");
			s.append("# of resigned games: " + this.numberOfResignedGames);
			s.append("\n");
			float avrg = (this.numberOfSolvedGames + this.numberOfUnsolvedGames) != 0 
					? this.numberOfStepsTotal/(this.numberOfSolvedGames + this.numberOfUnsolvedGames) 
					: 0;
			s.append("Avrg. # of steps: " + avrg);
			s.append("\n");
			s.append("Total time: " + this.totalTime/1000 + " s");
			s.append("\n");
			avrg = this.numberOfPlayedGames != 0 
					? this.totalTime/this.numberOfPlayedGames 
					: 0;
			s.append("Average time/game: " + avrg/1000 + " s");
			s.append("\n");
			avrg =  (this.numberOfStepsTotal != 0) ? this.totalTime/this.numberOfStepsTotal : 0;
			s.append("Average time/step: " + avrg/1000 + " s");
			s.append("\n");
			s.append("Step statistics:");
			s.append("\n\n");
			for (int i = 1; i < this.numOfGamesForStepNum.length; i++) {
				s.append(i +" steps: " + this.numOfGamesForStepNum[i]);
				s.append("\n");
			}
			s.append("+++++++++++++++\n");
			return s.toString();
		}
	}
	
	public BoardInfo getBoardInfo() {
		if (mBoardInfo == null)
			return new BoardInfo();
		return mBoardInfo;
	}
	
	public static class BoardInfo {
		public int numOfColors = 0;
		public int numOfPlaces = 0;
		public int numOfRows =0 ;
		public boolean withRepetition = false;;
	}
	
	public static class GameInfo {
		
		public float time;
		public int steps;
		public boolean solved;
		
		public GameInfo(float time, int steps, boolean solved) {
			this.time = time;
			this.steps = steps;
			this.solved = solved;
		}
	}
}
