package de.uni_passau.fim.strategies;

/**
 * Genes are guesses (see {@link Guess}) together with fitness, as is needed
 * for genetic algorithms.
 */
public class Gene extends Guess implements Comparable<Gene> {

	private int mFitness;
	private int mHashcode = 0;
	
	/**
	 * Creates a Gene with fitness 0 from the given Guess.
	 */
	public Gene(Guess g) {
		super(g.colors, g.numberOfColors);
		mFitness = 0;
	}
	
	/**
	 * Creates a new Gene with fitness 0 from the given color array <code>c</code>.
	 * @param c The initial color array.
	 * @param numberOfColors Total number of colors used in the current Mastermind.
	 */
	public Gene(int[] c, int numberOfColors) {
		super(c, numberOfColors);
		mFitness = 0;
	}
	
	/**
	 * Returns the current fitness of this gene.
	 */
	public int getFitness() {
		return mFitness;
	}
	
	/**
	 * Set given fitness.
	 * @param fitness
	 */
	public void setFitness(int fitness) {
		mFitness = fitness;
	}

	/**
	 * Two genes are compared according to their fitness.
	 * @param o1
	 * @param o2
	 * @return The difference fitness of <code>o1</code> minus fitness of <code>o2<code>.
	 */
	public int compare(Gene o1, Gene o2) {
		return o1.mFitness-o2.mFitness;
	}

	/** Returns true only if <code>o</code> is a gene containing the 
	 * same guess and fitness as this gene. */
	public boolean equals(Object o) {
		if (this == o) 
			return true;
		if (!(o instanceof Gene)) 
			return false;
		Gene g = (Gene)o;
		return super.equals(g) && this.getFitness() == g.getFitness();
	}
	
	/** Returns the current hash code. */
	public int hashCode() {
		if (mHashcode == 0) {
			mHashcode = 17 + super.hashCode();
		}
		return mHashcode;
	}

	/** See {@link Comparable#compareTo(Object)} */
	public int compareTo(Gene o) {
		return compare(this, o);
	}
	
	/** Returns a string in the form of <code>([2,2,3,3], 4)</code>,
	 * where <code>[2,2,3,3]</code> is the genes color array and 4 its fitness. */
	public String toString() {
		return "(" + super.toString() + ", " + mFitness + ")";
	}
	
}
